import logging
import os
from pathlib import Path

from pyaugmecon.options import Options

logging.getLogger("pyomo.core").setLevel(logging.ERROR)


class Logs:
    def __init__(self, opts: Options):
        """
        Initializes the Logs object.

        Parameters:
        ----------
        opts: Options
            An object containing the options of the program.

        """
        self.opts = opts

        # Create log directory if it does not exist
        if not os.path.exists(self.opts.logdir):
            os.makedirs(self.opts.logdir)

        # Set the log directory and log file path
        self.logdir = f"{Path().absolute()}/{self.opts.logdir}/"
        self.logfile = f"{self.logdir}{self.opts.log_name}.log"
        self.logger = logging.getLogger(opts.log_name)  # Set up logger object
        self.handler = logging.FileHandler(self.logfile)  # Set up handler object for logger object
        self.formatter = logging.Formatter("[%(asctime)s] %(message)s")  # Set up formatter object for handler object
        self.handler.setFormatter(self.formatter)  # Add formatter object to handler object
        self.logger.addHandler(self.handler)  # Add handler object to logger object
        self.handler.setLevel(logging.INFO)  # Set logging level for the handler object



import itertools
from multiprocessing import Manager

from pyaugmecon.options import Options


class Flag:
    def __init__(self, opts: Options):
        """
        Initialize the Flag class.

        Parameters
        ----------
        opts : Options
            An instance of the Options class.

        """
        self.opts = opts
        self.flag = Manager().dict() if self.opts.shared_flag else {}

    def set(self, flag_range, value, iter):
        """
        Set the specified value for all possible index combinations in the dictionary.

        Parameters
        ----------
        flag_range : function
            A function that returns the range of values that an index can take.
        value : any
            The value to be set.
        iter : iterable
            An iterable containing the indices.

        """
        indices = list(itertools.product(*(flag_range(o) for o in iter)))
        self.flag.update({gp: value for gp in indices})

    def get(self, i):
        """
        Get the value of the flag for the given index combination.

        Parameters
        ----------
        i : tuple
            The index combination.

        Returns
        -------
        any
            The value of the flag for the given index combination or 0 if it doesn't exist.

        """
        return self.flag.get(i, 0)



import numpy as np

from pyaugmecon import PyAugmecon
from tests.helper import Helper
from tests.optimization_models import two_objective_model

model_type = "two_objective_model"

options = {
    "name": model_type,
    "grid_points": 10,
}

py_augmecon = PyAugmecon(two_objective_model(), options)
py_augmecon.solve()


def test_payoff_table():
    payoff = np.array([[20, 160], [8, 184]])
    assert Helper.array_equal(py_augmecon.model.payoff, payoff, 2)


def test_pareto_sols():
    pareto_sols = np.array(
        [
            [8, 184],
            [9.33, 181.33],
            [10.67, 178.67],
            [12, 176],
            [13.33, 173.33],
            [14.67, 170.67],
            [16, 168],
            [17.33, 165.33],
            [18.67, 162.67],
            [20, 160],
        ]
    )
    assert Helper.array_equal(py_augmecon.unique_pareto_sols, pareto_sols, 2)



import logging
import queue
from multiprocessing import Manager, Queue

import numpy as np

from pyaugmecon.options import Options


class QueueHandler:
    def __init__(self, work, opts: Options):
        """
        Initialize a QueueHandler object.

        Parameters
        ----------
        work : Any
            The input work.
        opts : Options
            The options object.

        """
        self.work = work
        self.opts = opts
        self.logger = logging.getLogger(opts.log_name)
        self.result_q = Queue()
        self.proc_count = 0
        self.job_qs = []

    def get_longest_q(self):
        """
        Get the index of the job queue with the most items.

        Returns
        -------
        int or None
            The index of the job queue with the most items, or None if all queues are empty.

        """
        q_lengths = [q.qsize() for q in self.job_qs]
        if all(q == 0 for q in q_lengths):
            return None
        return np.argmax(q_lengths)

    def get_work(self, i):
        """
        Get work for the i-th process.

        Parameters
        ----------
        i : int
            Index of the process

        Returns
        -------
        list or None
            A list of items to process, or None if there is no more work.

        """
        try:
            return self.job_qs[i].get_nowait()  # Try to get the work for a given process without blocking
        except queue.Empty:  # If the queue is empty
            longest_q = self.get_longest_q()  # Get the index of the job queue with the most items
            if self.opts.redivide_work and longest_q is not None:
                # If the `redivide_work` flag is set to True and there is work available, redivide the work
                return self.get_work(longest_q)
            if self.opts.process_logging:
                # If process logging is enabled, log that the process exited
                self.logger.info(f"PID: {i} exited")
            self.result_q.put("STOP")  # Put "STOP" in the result queue to signal the end of processing
            return None

    def put_result(self, result):
        """
        Put a result in the result queue.

        Parameters
        ----------
        result : any
            The result to put in the queue.

        """
        self.result_q.put(result)  # Put a result in the result queue

    def get_result(self):
        """
        Get the results from the result queue.

        Returns
        -------
        list
            A list of results.

        """
        results = []  # List to hold the results
        for _ in range(self.proc_count):
            while True:
                result = self.result_q.get()
                if result == "STOP":
                    break  # If the result is "STOP", break out of the loop
                results.append(result)  # Otherwise, append the result to the list
        return results

    def empty_job_qs(self):
        """
        Empty the job queues.
        """
        for job_q in self.job_qs:
            while True:
                try:
                    job_q.get_nowait()
                except queue.Empty:
                    break  # Empty the job queues

    def split_work(self):
        """
        Split the input work into blocks and create a job queue for each process.
        """
        block_size = self.opts.gp  # Get the block size from the options
        blocks = [
            self.work[i : i + block_size] for i in range(0, len(self.work), block_size)
        ]  # Divide the work into blocks

        # Divide the blocks into sub-blocks and remove empty sub-blocks
        blocks = [x for x in np.array_split(np.array(blocks), self.opts.cpu_count) if x.size > 0]

        manager = Manager()
        self.proc_count = len(blocks)  # Set the number of processes to be used
        self.job_qs = [manager.Queue() for _ in range(self.proc_count)]  # Create a job queue for each process
        self.logger.info(f"Dividing grid over {self.proc_count} process(es)")  # Log the number of processes

        for i, block in enumerate(blocks):
            for item in block:
                item = [tuple(x) for x in item.tolist()]
                self.job_qs[i].put_nowait(item)  # Put each item in the job queue for the corresponding process



import logging
import os

import cloudpickle
import numpy as np
import pandas as pd
import pyomo.environ as pyo
from pyomo.core.base import (
    Any,
    ConstraintList,
    NonNegativeReals,
    Param,
    Set,
    Var,
    maximize,
    minimize,
)

from pyaugmecon.helper import Counter, ProgressBar
from pyaugmecon.options import Options


class Model:
    def __init__(self, model: pyo.ConcreteModel, opts: Options):
        """
        Initialize the Model class.

        Parameters
        ----------
        model : pyo.ConcreteModel
            Pyomo concrete model to solve.
        opts : Options
            Options object for the Pyomo model.

        """
        self.model = model
        self.opts = opts
        self.logger = logging.getLogger(opts.log_name)

        self.n_obj = len(model.obj_list)
        self.iter_obj = range(self.n_obj)
        self.iter_obj2 = range(self.n_obj - 1)

        # Setup progress bar
        self.to_solve = opts.gp ** (self.n_obj - 1) + self.n_obj**2
        self.progress = ProgressBar(Counter(), self.to_solve)

        self.models_solved = Counter()
        self.infeasibilities = Counter()

        if self.n_obj < 2:
            raise ValueError("Too few objective functions provided")

    def obj(self, i):
        """
        Return the i-th objective function.

        Parameters
        ----------
        i : int
            The index of the objective function to return.

        Returns
        -------
        obj : pyo.Objective
            The i-th objective function.

        """
        return self.model.obj_list[i + 1]

    def obj_val(self, i):
        """
        Return the value of the i-th objective function.

        Parameters
        ----------
        i : int
            The index of the objective function to return.

        Returns
        -------
        obj_val : float
            The value of the i-th objective function.

        """
        return self.obj(i)()

    def obj_expr(self, i):
        """
        Return the expression of the i-th objective function.

        Parameters
        ----------
        i : int
            The index of the objective function to return.

        Returns
        -------
        obj_expr : pyo.Expression
            The expression of the i-th objective function.

        """
        return self.obj(i).expr

    def obj_sense(self, i):
        """
        Return the sense of the i-th objective function.

        Parameters
        ----------
        i : int
            The index of the objective function to return.

        Returns
        -------
        obj_sense : pyo.ObjectiveSense
            The sense of the i-th objective function.

        """
        return self.obj(i).sense

    def slack_val(self, i):
        """
        Return the value of the slack variable of the i-th constraint.

        Parameters
        ----------
        i : int
            The index of the constraint.

        Returns
        -------
        slack_val : float
            The value of the slack variable of the i-th constraint.

        """
        return self.model.Slack[i + 1].value

    def obj_activate(self, i):
        """
        Activate the i-th objective function.
        """
        self.obj(i).activate()

    def obj_deactivate(self, i):
        """
        Deactivate the i-th objective function.
        """
        self.obj(i).deactivate()

    def solve(self):
        """
        Solve the model using the specified solver.

        The result, termination condition, and solver status are stored as class attributes.
        """
        opt = pyo.SolverFactory(self.opts.solver_name, solver_io=self.opts.solver_io, manage_env=True)
        opt.options.update(self.opts.solver_opts)
        try:
            self.result = opt.solve(self.model)
            self.term = self.result.solver.termination_condition
            self.status = self.result.solver.status
        finally:
            if self.opts.solver_name.lower() == "gurobi":
                opt.close()

    def get_vars(self):
        """
        Return a dictionary containing variable names and values extracted from the Pyomo model.

        Returns
        -------
        vars_dict : dict
            A dictionary containing variable names as keys and Pandas Series containing the variable values as values.
        """
        model_vars = self.model.component_map(ctype=Var, active=True)
        vars_dict = {
            v.name: pd.Series(v.extract_values(), index=v.extract_values().keys()) for v in model_vars.values()
        }
        return vars_dict

    def pickle(self):
        """
        Pickle the Pyomo model to a file.
        """
        model_file = open(self.opts.model_fn, "wb")
        cloudpickle.dump(self.model, model_file)
        del self.model

    def unpickle(self):
        """
        Unpickle the Pyomo model from a file.
        """
        model_file = open(self.opts.model_fn, "rb")
        self.model = cloudpickle.load(model_file)

    def clean(self):
        """
        Remove the Pyomo model file.
        """
        if os.path.exists(self.opts.model_fn):
            os.remove(self.opts.model_fn)

    def is_optimal(self):
        """
        Check if the Pyomo model has been solved optimally.

        Returns
        -------
        bool
            True if the Pyomo model has been solved optimally, False otherwise.
        """
        return self.status == pyo.SolverStatus.ok and self.term == pyo.TerminationCondition.optimal

    def is_infeasible(self):
        """
        Check if the Pyomo model is infeasible.

        Returns
        -------
        bool
            True if the Pyomo model is infeasible, False otherwise.
        """
        return (
            self.term == pyo.TerminationCondition.infeasible
            or self.term == pyo.TerminationCondition.infeasibleOrUnbounded
        )

    def min_to_max(self):
        """
        Convert all minimize objectives to maximize objectives and negate their expressions.

        This method modifies the Pyomo model in-place.

        """
        # Determine objective sense for each objective
        self.obj_goal = [-1 if self.obj_sense(o) == minimize else 1 for o in self.iter_obj]

        # Cconvert minimize objectives to maximize objectives and negate their expressions
        for o in self.iter_obj:
            if self.obj_sense(o) == minimize:
                self.model.obj_list[o + 1].sense = maximize
                self.model.obj_list[o + 1].expr = -self.model.obj_list[o + 1].expr

    def construct_payoff(self):
        """
        Construct a payoff matrix for all pairs of objective functions.

        The payoff matrix is filled with the optimal objective values achieved when optimizing each pair of objectives.

        """
        self.logger.info("Constructing payoff")
        self.progress.set_message("constructing payoff")

        def set_payoff(i, j):
            """
            Helper function that optimizes the Pyomo model with objective function j and saves its value in the payoff
            matrix.

            Parameters
            ----------
            i : int
                The index of the first objective function to use as a constraint.
            j : int
                The index of the second objective function to optimize.

            """
            self.obj_activate(j)
            self.solve()
            self.progress.increment()
            self.payoff[i, j] = self.obj_val(j)
            self.obj_deactivate(j)

        # Initialize payoff matrix with infinity values
        self.payoff = np.full((self.n_obj, self.n_obj), np.inf)
        self.model.pcon_list = ConstraintList()

        # Optimize each objective function independently (diagonal elements)
        for i in self.iter_obj:
            set_payoff(i, i)

        # Optimize each pair of objective functions (off-diagonal elements)
        for i in self.iter_obj:
            self.model.pcon_list.add(expr=self.obj_expr(i) == self.payoff[i, i])

            for j in self.iter_obj:
                if i != j:
                    set_payoff(i, j)
                    self.model.pcon_list.add(expr=self.obj_expr(j) == self.payoff[i, j])

            self.model.pcon_list.clear()

    def find_obj_range(self):
        """
        Find the range of each objective function and create a grid of points that are used as constraints.

        This method modifies the Pyomo model in-place and sets the following class attributes:
            - e : ndarray
                A 2D array containing the gridpoints of p-1 objective functions that are used as constraints.
            - obj_range : ndarray
                A 1D array containing the range of each objective function.

        """
        self.logger.info("Finding objective function range")

        # Initialize gridpoints and objective range arrays
        self.e = np.zeros((self.n_obj - 1, self.opts.gp))
        self.obj_range = np.zeros(self.n_obj - 1)

        # Find range of each objective function
        for i in self.iter_obj2:
            if self.opts.nadir_p:
                obj_min = self.opts.nadir_p[i]
            else:
                obj_min = self.opts.nadir_r * np.min(self.payoff[:, i + 1])

            obj_max = np.max(self.payoff[:, i + 1])
            self.obj_range[i] = obj_max - obj_min
            self.e[i] = np.linspace(obj_min, obj_max, self.opts.gp)

    def convert_prob(self):
        """
        Convert the multi-objective optimization problem to a single-objective optimization problem with constraints.

        This method modifies the Pyomo model in-place and adds the following variables and constraints:
            - Os : Set
                A Pyomo set that contains the indices of the objective functions that are used as constraints.
            - Slack : Var
                A Pyomo variable that represents the slack for each objective function used as a constraint.
            - e : Param
                A Pyomo parameter that contains the gridpoints of the objective functions that are used as constraints.
            - con_list : ConstraintList
                A Pyomo constraint list that contains the constraints for the objective functions.

        """
        self.logger.info("Converting optimization problem")

        # Create constraint list
        self.model.con_list = ConstraintList()

        # Create set of objective functions and Slack variable for each function
        self.model.Os = Set(ordered=True, initialize=[o + 2 for o in self.iter_obj2])
        self.model.Slack = Var(self.model.Os, within=NonNegativeReals)

        # Create parameter for gridpoints of objective functions used as constraints
        self.model.e = Param(self.model.Os, within=Any, mutable=True)

        # Add objective functions as constraints with slack variables
        for o in range(1, self.n_obj):
            self.model.obj_list[1].expr += self.opts.eps * (
                10 ** (-o + 1) * self.model.Slack[o + 1] / self.obj_range[o - 1]
            )

            self.model.con_list.add(
                expr=self.model.obj_list[o + 1].expr - self.model.Slack[o + 1] == self.model.e[o + 1]
            )



import time
from multiprocessing import Lock, Value


class Helper:
    """A class of helper functions."""

    @staticmethod
    def clear_line():
        """Clears the current line in the console."""
        print(" " * 80, end="\r", flush=True)

    @staticmethod
    def separator():
        """Returns a string of '=' characters for use as a separator."""
        return "=" * 30

    @staticmethod
    def keys_to_list(d: dict):
        """Converts the keys of a dictionary to a list."""
        return [list(key) for key in d.keys()]


class Counter:
    """A thread-safe counter class."""

    def __init__(self, init_val: int = 0):
        """Initializes the counter with an optional initial value."""
        self.val = Value("i", init_val)
        self.lock = Lock()

    def increment(self):
        """Increments the counter."""
        with self.lock:
            self.val.value += 1

    def value(self):
        """Returns the current value of the counter."""
        with self.lock:
            return self.val.value


class Timer:
    """A class for measuring elapsed time."""

    def __init__(self):
        """Starts the timer."""
        self.start_time = time.time()

    def get(self):
        """Returns the elapsed time since the timer was started."""
        return time.time() - self.start_time


class ProgressBar:
    """A class for displaying a progress bar."""

    def __init__(self, counter: Counter, total: int, init_message: str = ""):
        """Initializes the progress bar with a counter and a total number of iterations."""
        self.counter = counter
        self.total = total
        self.message = init_message
        self.bar = ""

    def set_message(self, message):
        """Sets the message displayed next to the progress bar."""
        self.message = message
        Helper.clear_line()
        self.print(True)

    def print(self, force):
        """Prints the progress bar."""
        bar_len = 40

        progress = self.counter.value() / float(self.total)
        filled_len = int(round(bar_len * progress, 1))
        percents = round(100.0 * progress, 1)
        bar = "=" * filled_len + "-" * (bar_len - filled_len)

        if self.bar != bar or force:
            self.bar = bar
            print(f"[{bar}] {percents}% ... ({self.message})", end="\r", flush=True)

    def increment(self):
        """Increments the counter and updates the progress bar."""
        self.counter.increment()
        self.print(False)



import logging
from multiprocessing import Process

from pyaugmecon.flag import Flag
from pyaugmecon.model import Model
from pyaugmecon.options import Options
from pyaugmecon.queue_handler import QueueHandler


class SolverProcess(Process):
    def __init__(self, p_num, opts: Options, model: Model, queues: QueueHandler, flag: Flag):
        """
        Initialize the SolverProcess object.

        Parameters
        ----------
        p_num : int
            The process number.
        opts : Options
            The options object.
        model : Model
            The model object.
        queues : QueueHandler
            The queue handler object.
        flag : Flag
            The flag object.

        """
        super().__init__()
        self.p_num = p_num
        self.opts = opts
        self.model = model
        self.queues = queues
        self.flag = flag
        self.logger = None

    def run(self):
        """
        Run the SolverProcess.

        This function runs the SolverProcess. It retrieves work items from the queue and processes them until there is
        no more work to do. For each work item, it sets up the model and solves it. If the `early_exit` flag is set to
        True and the model is infeasible, the function sets the flag and jumps to the next work item. If the `bypass`
        flag is set to True and the model is optimal, the function sets the flag and jumps to the next work item. If
        the model is optimal, the function calculates the solutions and puts them in the result queue.

        """
        jump = 0  # Initialize variable 'jump' to zero

        # Check if process logging is enabled
        if self.opts.process_logging:
            # If enabled, initialize logger with log name provided in 'self.opts.log_name'
            self.logger = logging.getLogger(self.opts.log_name)
            self.logger.setLevel(logging.INFO)  # Set log level to INFO

        # Load pickled model
        self.model.unpickle()

        # Run indefinitely until there's no more work
        while True:
            # Get work from queue for this process number
            work = self.queues.get_work(self.p_num)

            # If no more work, break the loop
            if not work:
                break

            # Process each job in the work list
            for c in work:
                log = f"Process: {self.p_num}, index: {c}, "  # Initialize logging string
                cp_end = self.opts.gp - 1  # Set the end index for this job
                self.model.progress.increment()  # Increment progress counter for this model

                # Define helper functions to handle jump, bypass, and early exit scenarios
                def do_jump(i, jump):
                    return min(jump, abs(cp_end - i))

                def bypass_range(i):
                    if i == 0:
                        return range(c[i], c[i] + 1)
                    else:
                        return range(c[i], c[i] + b[i] + 1)

                def early_exit_range(i):
                    if i == 0:
                        return range(c[i], c[i] + 1)
                    else:
                        return range(c[i], cp_end)

                # Check if flag is enabled and if the current job has a flag set
                if self.opts.flag and self.flag.get(c) != 0 and jump == 0:
                    # If jump is not set and there's a flag for this job, set jump
                    jump = do_jump(c[0] - 1, self.flag.get(c))

                # If jump is set, skip to the next iteration of the loop
                if jump > 0:
                    jump = jump - 1
                    continue

                # Log model progress for each objective
                for o in self.model.iter_obj2:
                    log += f"e{o + 1}: {self.model.e[o, c[o]]}, "
                    self.model.model.e[o + 2] = self.model.e[o, c[o]]

                # Activate objective 0 and solve the model
                self.model.obj_activate(0)
                self.model.solve()
                self.model.models_solved.increment()

                # Check if early exit is enabled and if the model is infeasible
                if self.opts.early_exit and self.model.is_infeasible():
                    # If so, increment infeasibilities counter
                    self.model.infeasibilities.increment()

                    # Set flag if flag is enabled
                    if self.opts.flag:
                        self.flag.set(early_exit_range, self.opts.gp, self.model.iter_obj2)

                    jump = do_jump(c[0], self.opts.gp)  # Set jump
                    log += "infeasible"  # Log infeasibility

                    # Log progress if process logging is enabled
                    if self.opts.process_logging:
                        self.logger.info(log)

                    continue  # Skip to next iteration of loop

                # Calculate slack values and set jump if bypass is enabled
                elif self.opts.bypass and self.model.is_optimal():
                    b = []
                    for i in self.model.iter_obj2:
                        step = self.model.obj_range[i] / (self.opts.gp - 1)
                        slack = round(self.model.slack_val(i + 1))
                        b.append(int(slack / step))

                    # Log jump and set flag if enabled
                    log += f"jump: {b[0]}, "
                    if self.opts.flag:
                        self.flag.set(bypass_range, b[0] + 1, self.model.iter_obj2)
                    jump = do_jump(c[0], b[0])

                # If model is optimal, calculate and log solutions
                sols = []
                if self.model.is_optimal():
                    sols.append(
                        self.model.obj_val(0)
                        - self.opts.eps
                        * sum(
                            10 ** (-1 * (o)) * self.model.slack_val(o + 1) / self.model.obj_range[o]
                            for o in self.model.iter_obj2
                        )
                    )

                    for o in self.model.iter_obj2:
                        sols.append(self.model.obj_val(o + 1))

                    # Put results into queue as a dictionary
                    sols_dict = {tuple(sols): self.model.get_vars()}
                    self.queues.put_result(sols_dict)

                    # Log solutions if process logging is enabled
                    log += f"solutions: {sols}"
                    if self.opts.process_logging:
                        self.logger.info(log)
